from base64 import urlsafe_b64encode
from hashlib import sha256
from json import dumps
import os
import aiohttp
from mighty_sdk_core.auth.secret import random_secret
from mighty_sdk_core.auth.types import GenerateCodeVerifier, GrantType, OAuthAuthorizationParam, OAuthTokenParam, Token, BiscuitToken
import urllib
from dotenv import load_dotenv

from mighty_sdk_core.data.types import EncryptedSharedData


API_KEY_HEADER = "x-api-key"
BISCUIT_HEADER = "x-biscuit"
AUTHORIZATION_HEADER = "Authorization"
SECRET_LENGTH = 128

base_headers = {
    "Content-Type": "application/json",
    "User-Agent": "Mozilla/5.0"
}

# Load environment variables from .env file
load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

mighty_base_url = os.getenv("MIGHTY_BASE_URL", "http://localhost:8080")
authorize_url = f"{mighty_base_url}/mighty-oauth/authorization"
token_url = f"{mighty_base_url}/mighty-oauth/token"
refresh_token_url = f"{mighty_base_url}/mighty-oauth/refresh-token"
user_info_url = f"{mighty_base_url}/api/v1/app/user-data"
user_info_url_biscuit = f"{mighty_base_url}/api/v1/app/biscuit/user-data"
biscuit_token_url = f"{mighty_base_url}/mighty-oauth/biscuit-token"

def generate_code_verifier() -> GenerateCodeVerifier:
    """
    Generates a code verifier for the OAuth2 authorization process.

    A code verifier is a cryptographically random string that is used to
    mitigate authorization code interception attacks. It is part of the
    Proof Key for Code Exchange (PKCE) extension to OAuth2.

    Returns:
        str: The generated code verifier.
    """
    # Random the code verifier
    # Reference: https://aaronparecki.com/oauth-2-simplified/#:~:text=(ietf.org).-,Authorization,-Create%20a%20random
    code_verifier = random_secret(SECRET_LENGTH)

    # Hash and encode the code verifier
    bin_code_verifier = sha256(code_verifier.encode('utf-8')).digest()
    code_challenge = urlsafe_b64encode(bin_code_verifier).decode("utf-8").replace("=", "")

    return GenerateCodeVerifier(code_verifier=code_verifier, code_challenge=code_challenge)

def get_authorization_url(oauth_config: OAuthAuthorizationParam) -> str:
    """
    Builds the authorization URL for the application.

    This method constructs the URL that can be used to authorize the application
    with the necessary permissions and access rights.

    Returns:
        str: The constructed authorization URL.
    """

    params = urllib.parse.urlencode(oauth_config.model_dump())
    return f"{authorize_url}?{params}"

async def exchange_code_for_token(code: str, application_api_key: str, oauth_config: OAuthTokenParam) -> Token:
    """
    Exchanges an authorization code for an access token.

    This method sends the authorization code to the token endpoint to obtain an access token.

    Args:
        code (str): The authorization code to exchange for an access token.
        application_api_key (str): The API key of the application.

    Returns:
        Token: The access token and associated information.
    """

    async with aiohttp.ClientSession() as session:
        headers = {
            **base_headers,
            API_KEY_HEADER: application_api_key
        }
        payload = {
            "code": code,
            "grant_type": GrantType.AUTHORIZATION_CODE.value,
            "code_verifier": oauth_config.code_verifier,
            "redirect_uri": oauth_config.redirect_uri,
            "client_id": oauth_config.client_id
        }
        async with session.post(token_url, data=dumps(payload), headers=headers) as response:
            try:
                response_json = await response.json()
                response_json["application_id"] = response_json.pop("username", None)
                return Token(**response_json)
            except Exception as e:
                raise Exception(f"Failed to exchange code for token: {e}")

async def get_user_info(token: str, application_api_key: str) -> EncryptedSharedData:
    """
    Retrieves user information from the Mighty API.

    This method sends a request to the user info endpoint with the provided access token
    to retrieve user-specific data.
    """
    async with aiohttp.ClientSession() as session:
        headers = {
            **base_headers,
            AUTHORIZATION_HEADER: f"Bearer {token}",
            API_KEY_HEADER: application_api_key
        }
        async with session.get(user_info_url, headers=headers) as response:
            try:
                return EncryptedSharedData(**await response.json())
            except Exception as e:
                raise Exception(f"Failed to get user info: {e}. Please verify that the application has invited you and that your access token is valid. Refer to the Mighty Network API documentation at docs.mightynetwork.ai for further guidance.")

async def exchange_code_for_biscuit_token(code: str, expiration: int, usage_once: bool, application_api_key: str, oauth_config: OAuthTokenParam) -> BiscuitToken:
    """
    Exchanges an authorization code for an access token.
    This method sends the authorization code to the token endpoint to obtain an access token.
    Args:
        code (str): The authorization code to exchange for an access token.
        application_api_key (str): The API key of the application.
    Returns:
        Token: The access token and associated information.
    """
    async with aiohttp.ClientSession() as session:
        headers = {
            **base_headers,
            API_KEY_HEADER: application_api_key
        }
        payload = {
            "code": code,
            "grant_type": GrantType.AUTHORIZATION_CODE.value,
            "code_verifier": oauth_config.code_verifier,
            "redirect_uri": oauth_config.redirect_uri,
            "client_id": oauth_config.client_id,
            "expiration": expiration,
            "usage_once": usage_once
        }
        async with session.post(biscuit_token_url, data=dumps(payload), headers=headers) as response:
            try:
                response_json = await response.json()
                return BiscuitToken(**response_json)
            except Exception as e:
                raise Exception(f"Failed to exchange code for token: {e}")
            
async def get_user_info_biscuit(biscuit: str, application_api_key: str) -> EncryptedSharedData:
    """
    Retrieves user information from the Mighty API.
    This method sends a request to the user info endpoint with the provided access token
    to retrieve user-specific data.
    """
    async with aiohttp.ClientSession() as session:
        headers = {
            **base_headers,
            BISCUIT_HEADER: biscuit,
            API_KEY_HEADER: application_api_key
        }
        async with session.get(user_info_url_biscuit, headers=headers) as response:
            try:
                return EncryptedSharedData(**await response.json())
            except Exception as e:
                raise Exception(f"Failed to get user info: {e}. Please verify that the application has invited you and that your access token is valid. Refer to the Mighty Network API documentation at docs.mightynetwork.ai for further guidance.")

async def refresh_token(refresh_token: str, application_api_key: str) -> Token:
    """
    Refreshes an access token using a refresh token.

    This method sends a request to the token endpoint with the provided refresh token
    to obtain a new access token.
    """
    async with aiohttp.ClientSession() as session:
        headers = {
            **base_headers,
            API_KEY_HEADER: application_api_key
        }
        payload = {
            "grant_type": GrantType.REFRESH_TOKEN.value,
            "refresh_token": refresh_token
        }
        async with session.post(refresh_token_url, data=dumps(payload), headers=headers) as response:
            try:
                response_json = await response.json()
                response_json["application_id"] = response_json.pop("username", None)
                return Token(**response_json)
            except Exception as e:
                raise Exception(f"Failed to refresh token: {e}")



import os
import aiohttp
from mighty_sdk_core.data.types import EncryptedSharedData
from mighty_sdk_core.data.decrypt import decrypt_shared_data
from mighty_sdk_core.mighty.client import MightyClient
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

mighty_base_url = os.getenv("MIGHTY_BASE_URL", "http://localhost:8080")
user_data_url = f"{mighty_base_url}/api/v1/encrypted-data"

"""
MightyUserDataClient is a client for fetching and decrypting user data from the Mighty Network API.

Attributes:
    api_key (str): The API key for accessing the Mighty Network API.
    api_public_key (str): The public key for decrypting the data.
    api_private_key (str): The private key for decrypting the data.
    data_url (str): The URL endpoint for fetching the encrypted data.

Methods:
    get_data() -> str:
        Asynchronously fetches and decrypts the user data from the API.
"""
class MightyUserDataClient(MightyClient):
    api_key: str
    api_public_key: str
    api_private_key: str

    def __init__(self, api_key: str, api_public_key: str, api_private_key: str):
        super().__init__()
        self.api_key = api_key
        self.api_public_key = api_public_key
        self.api_private_key = api_private_key

    async def get_data(self) -> str:
        try:
            encrypted_data = await self._get_data_by_api_key(self.api_key, user_data_url)
            decrypted_data = decrypt_shared_data(encrypted_data, self.api_public_key, self.api_private_key)

            return decrypted_data
        except Exception as e:
            raise e

    async def _get_data_by_api_key(self, api_key: str, data_url: str) -> EncryptedSharedData:
        async with aiohttp.ClientSession() as session:
            headers = {
                "x-api-key": api_key,
                "User-Agent": "Mozilla/5.0"
            }
            async with session.get(data_url, headers=headers) as response:
                return EncryptedSharedData(**await response.json())


import os
from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from semver_range import Version
from .aad_meta import AADMeta
from .json_datagram import JSONDatagram
from .utils import concat_uint8_arrays

NONCE_LENGTH = 12

class TaggedSecretBox:
    """
    A class to handle encryption and decryption using the ChaCha20Poly1305 algorithm
    with additional authenticated data (AAD) metadata.

    Attributes:
        key (ChaCha20Poly1305): The encryption key.
    """
    def __init__(self, key_bytes: bytes) -> None:
        """
        Initialize a TaggedSecretBox instance.

        :param key_bytes: The encryption key as bytes.
        """
        self.key = ChaCha20Poly1305(key_bytes)

    def encrypt(self, datagram: JSONDatagram, data: str, nonce = os.urandom(NONCE_LENGTH)) -> bytearray:
        """
        Encrypt the given data using the provided JSON datagram and nonce.

        :param datagram: The JSONDatagram object containing metadata for encryption.
        :param data: The plaintext data to be encrypted.
        :param nonce: The nonce to be used for encryption. Defaults to a random nonce.
        :return: The encrypted data as a bytearray.
        """
        aad = AADMeta(datagram.version, datagram.type, nonce)
        aad_serialized = aad.serialize()

        return concat_uint8_arrays(aad_serialized, self.key.encrypt(nonce, datagram.serialize(data), aad_serialized))

    def decrypt(self, datagram: JSONDatagram, bytes: bytes) -> str:
        """
        Decrypt the given byte array using the provided JSON datagram.

        :param datagram: The JSONDatagram object containing metadata for decryption.
        :param bytes: The encrypted data as a byte array.
        :return: The decrypted data as a string.
        :raises Exception: If decryption fails due to invalid key, type mismatch, or version constraint violation.
        """
        header = AADMeta.deserialize(bytes)
        if header is None or "metadata" not in header.keys():
            raise Exception("Couldn't decrypt: no header in provided data")
        
        decrypted = self.key.decrypt(header["metadata"].nonce, header["content"], header["raw_metadata"])

        if decrypted is None:
            raise Exception("Couldn't decrypt: invalid key")

        if datagram.type != header["metadata"].type:
            raise Exception("Couldn't decrypt: encrypted type ({}) doesn't match datagram type ({})".format(header["metadata"].type, datagram.type))

        if not (Version(header["metadata"].version) in datagram.version_constraint):
            raise Exception("Couldn't decrypt: encrypted version ({}) doesn't match datagram version constraint ({})".format(header["metadata"].version, datagram.version_constraint))

        return datagram.deserialize(decrypted)


from enum import Enum
from pydantic import BaseModel, ConfigDict, Field


class CodeChallengeMethod(str, Enum):
    SHA256 = 'sha256'

class ResponseType(str, Enum):
    CODE = "code"

class GrantType(str, Enum):
    AUTHORIZATION_CODE = "authorization_code"
    REFRESH_TOKEN = "refresh_token"
    
class GenerateCodeVerifier(BaseModel):
    code_verifier: str
    code_challenge: str

class OAuthAuthorizationParam(BaseModel):
    client_id: str
    redirect_uri: str
    state: str
    code_challenge: str
    code_challenge_method: CodeChallengeMethod = Field(default=CodeChallengeMethod.SHA256)
    response_type: ResponseType = Field(default=ResponseType.CODE)

    model_config = ConfigDict(
        use_enum_values=True,
        validate_default=True
    )

class OAuthTokenParam(BaseModel):
    code_verifier: str
    client_id: str
    redirect_uri: str

class Token(BaseModel):
    application_id: str
    roles: list[str] = Field(default_factory=list)
    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int



from enum import Enum

from pydantic import BaseModel

class ApplicationInformation(BaseModel):
    name: str
    description: str
    website: str
    is_active: bool
    permissions: list[str] # TODO: This type mighty change



import os
from mighty_sdk_core.auth.oauth import get_user_info, get_user_info_biscuit
from mighty_sdk_core.data.decrypt import decrypt_shared_data
from mighty_sdk_core.mighty.client import MightyClient
from mighty_sdk_core.mighty.types import ApplicationInformation


class MightyApplicationClient(MightyClient):

    def __init__(self, api_key: str, app_private_key: str):
        self.api_key = api_key
        self.app_private_key = app_private_key

    def get_application_information(self) -> ApplicationInformation:
        pass

    async def get_user_data(self, token: str) -> str:
        """
        Fetches and decrypts user data using the provided token.

        Args:
            token (str): The token used for authentication.

        Returns:
            str: The decrypted user data.

        Raises:
            Exception: If there is an error during the fetching or decryption process.
        """
        # Get the user's shared encrypted data
        encrypted_data = await get_user_info(token, self.api_key)

        # Decrypt the shared data
        decrypted_data = decrypt_shared_data(encrypted_data, encrypted_data.associated_public_key, self.app_private_key)

        return decrypted_data
    
    async def get_user_data_biscuit(self, token: str) -> str:
        """
        Fetches and decrypts user data using the provided token.

        Args:
            token (str): The token used for authentication.

        Returns:
            str: The decrypted user data.

        Raises:
            Exception: If there is an error during the fetching or decryption process.
        """
        # Get the user's shared encrypted data
        encrypted_data = await get_user_info_biscuit(token, self.api_key)

        # Decrypt the shared data
        decrypted_data = decrypt_shared_data(encrypted_data, encrypted_data.associated_public_key, self.app_private_key)

        return decrypted_data




from mighty_sdk_core.crypto.json_datagram import JSONDatagram
from mighty_sdk_core.crypto.symmetric_encryption import decrypt_symmetric
from mighty_sdk_core.data.types import EncryptedSharedData
from mighty_sdk_core.crypto.asymmetric_encryption import string_decrypt_asymmetric


SED_JSON_DATAGRAM = JSONDatagram("User Shared Encrypted Data Datagram")

def decrypt_shared_data(encrypted_data: EncryptedSharedData, public_key: str, private_key: str) -> str:
    """
    Decrypts the shared data using the provided public and private keys.

    Args:
        encrypted_data (EncryptedSharedData): The encrypted shared data to decrypt.
        public_key (str): The public key for decryption.
        private_key (str): The private key for decryption.

    Returns:
        str: The decrypted shared data.
    """
    
    data_key = string_decrypt_asymmetric(
        private_key,
        public_key,
        encrypted_data.public_key_encrypted_key
    )

    return decrypt_symmetric(
        message = encrypted_data.encrypted_data,
        symmetric_key = data_key,
        datagram = SED_JSON_DATAGRAM
    )



import base64
from .tagged_secret_box import *

def raw_encrypt_symmetric(content: str, symmetric_key: str, datagram: JSONDatagram):
    """
    Encrypts the given content using a symmetric key and a JSON datagram.

    :param content: The plaintext content to be encrypted.
    :param symmetric_key: The base64 encoded symmetric key for encryption.
    :param datagram: The JSONDatagram object containing metadata for encryption.
    :return: The encrypted content as a byte array.
    """
    secret_box = TaggedSecretBox(base64.b64decode(symmetric_key))
    return secret_box.encrypt(datagram, content)

def encrypt_symmetric(content: str, symmetric_key: str, datagram: JSONDatagram):
    """
    Encrypts the given content using a symmetric key and a JSON datagram, then encodes the result in base64.

    :param content: The plaintext content to be encrypted.
    :param symmetric_key: The base64 encoded symmetric key for encryption.
    :param datagram: The JSONDatagram object containing metadata for encryption.
    :return: The encrypted content as a base64 encoded byte array.
    """
    return base64.b64encode(raw_encrypt_symmetric(content, symmetric_key, datagram))

def raw_decrypt_symmetric(message: bytes, symmetric_key: str, datagram: JSONDatagram):
    """
    Decrypts the given encrypted message using a symmetric key and a JSON datagram.

    :param message: The encrypted message as a byte array.
    :param symmetric_key: The base64 encoded symmetric key for decryption.
    :param datagram: The JSONDatagram object containing metadata for decryption.
    :return: The decrypted content as a string.
    """
    secret_box = TaggedSecretBox(base64.b64decode(symmetric_key))
    return secret_box.decrypt(datagram, message)

def decrypt_symmetric(message: str, symmetric_key: str, datagram: JSONDatagram):
    """
    Decrypts the given base64 encoded encrypted message using a symmetric key and a JSON datagram.

    :param message: The encrypted message as a base64 encoded string.
    :param symmetric_key: The base64 encoded symmetric key for decryption.
    :param datagram: The JSONDatagram object containing metadata for decryption.
    :return: The decrypted content as a string.
    """
    return raw_decrypt_symmetric(base64.b64decode(message), symmetric_key, datagram)



import base64
import nacl.secret
import nacl.utils
import nacl.public
import base64
from nacl.public import PrivateKey, PublicKey, Box
from .utils import utf8_bytes_to_string

def new_nonce():
    """
    Generate a new nonce for encryption.

    :return: A random nonce of size suitable for nacl.public.Box.
    """
    return nacl.utils.random(nacl.public.Box.NONCE_SIZE)

def decrypt_asymmetric(secret_or_shared_key, message_with_nonce):
    """
    Decrypt a message using a shared secret key and a nonce.

    :param secret_or_shared_key: The shared secret key for decryption.
    :param message_with_nonce: The encrypted message with nonce, base64 encoded.
    :return: The decrypted message as a string.
    :raises ValueError: If the message cannot be decrypted.
    """
    message_with_nonce_as_byte_array = base64.b64decode(message_with_nonce)
    nonce = message_with_nonce_as_byte_array[:nacl.secret.SecretBox.NONCE_SIZE]
    message = message_with_nonce_as_byte_array[nacl.secret.SecretBox.NONCE_SIZE:]
    
    decrypted = nacl.secret.SecretBox(secret_or_shared_key).decrypt(message, nonce)
    
    if not decrypted:
        raise ValueError("Could not decrypt message")
    
    decrypted_message = utf8_bytes_to_string(decrypted)
    return decrypted_message

def string_decrypt_asymmetric(my_private_key, their_public_key, encrypted_text):
    """
    Decrypt a base64 encoded message using asymmetric encryption.

    :param my_private_key: Base64 encoded private key of the recipient.
    :param their_public_key: Base64 encoded public key of the sender.
    :param encrypted_text: The encrypted message to be decrypted.
    :return: The decrypted message as a string.
    :raises Exception: If decryption fails.
    """
    try:
        my_private_key_bytes = base64.b64decode(my_private_key)
        their_public_key_bytes = base64.b64decode(their_public_key)
        my_private_key = PrivateKey(my_private_key_bytes)
        their_public_key = PublicKey(their_public_key_bytes)
        
        box = Box(my_private_key, their_public_key)
        
        shared_key = box.shared_key()
        
        return decrypt_asymmetric(shared_key, encrypted_text)
    except Exception as e:
        raise e
    
def encrypt_asymmetric(secret_or_shared_key, msg_str):
    """
    Encrypt a message using a shared secret key.

    :param secret_or_shared_key: The shared secret key for encryption.
    :param msg_str: The message to be encrypted as a string.
    :return: The encrypted message with nonce, base64 encoded.
    """
    # Convert message to bytes
    message = msg_str.encode('utf-8')
    
    # Generate a nonce
    nonce = new_nonce()
    
    box = nacl.secret.SecretBox(secret_or_shared_key)
    encrypted_message = box.encrypt(message, nonce)
    
    # Concatenate nonce and encrypted message
    full_message = nonce + encrypted_message.ciphertext
    
    # Encode the result as base64
    encoded_message = base64.b64encode(full_message).decode('utf-8')
    print(f"encoded_message {encoded_message}")
    
    return encoded_message

def string_encrypt_asymmetric(my_private_key, their_public_key, plain_text):
    """
    Encrypt a message using asymmetric encryption.

    :param my_private_key: Base64 encoded private key of the sender.
    :param their_public_key: Base64 encoded public key of the recipient.
    :param plain_text: The message to be encrypted as a string.
    """
    try:
        print(f"my_private_key {my_private_key}")
        print(f"their_public_key {their_public_key}")
        my_private_key_bytes = base64.b64decode(my_private_key)
        their_public_key_bytes = base64.b64decode(their_public_key)
        my_private_key = PrivateKey(my_private_key_bytes)
        their_public_key = PublicKey(their_public_key_bytes)
        
        box = Box(my_private_key, their_public_key)
        
        shared_key = box.shared_key()
        
        encrypt_asymmetric(shared_key, plain_text)
    except Exception as e:
        print(f"Error: {e}")




import varint
from typing import List

def utf8_string_to_bytes(data: str) -> bytes:
    """
    Convert a UTF-8 string to a byte array.

    :param data: The input string.
    :return: The encoded byte array.
    """
    return data.encode(encoding='utf-8')

def utf8_bytes_to_string(data: bytes) -> str:
    """
    Convert a byte array to a UTF-8 string.

    :param data: The input byte array.
    :return: The decoded string.
    """
    return data.decode('utf-8')

def concat_uint8_arrays(*u8s: List[bytes]) -> bytearray:
    """
    Concatenate multiple byte arrays into a single byte array.

    :param u8s: A variable number of byte arrays.
    :return: A single concatenated byte array.
    """
    total_len = sum(len(u8) for u8 in u8s)

    result = bytearray(total_len)

    index = 0
    for u8 in u8s:
        result[index:index+len(u8)] = u8
        index += len(u8)

    return result

def varint_prefixed(data: bytes) -> bytearray:
    """
    Prefix the given byte array with its length encoded as a varint.

    :param data: The input byte array.
    :return: A new byte array prefixed with the varint-encoded length.
    """
    length_encoded = varint.encode(len(data))
    return concat_uint8_arrays(length_encoded, data)

def extract_varint_prefixed(o): 
    """
    Extract a varint-prefixed chunk from the given object.

    :param o: A dictionary containing a key "bs" with byte data.
    :return: The extracted chunk as a byte array.
    """
    chunk_len = varint.decode_bytes(o["bs"])
    chunk_len_len = encoding_length(chunk_len)
    chunk = o["bs"][chunk_len_len:chunk_len_len + chunk_len]

    o["bs"] = o["bs"][chunk_len_len + chunk_len:]

    return chunk

def encoding_length(value):
    """
    Determine the number of bytes required to encode a given integer as a varint.

    :param value: The integer value to encode.
    :return: The number of bytes required for the varint encoding.
    """
    N1 = 2 ** 7
    N2 = 2 ** 14
    N3 = 2 ** 21
    N4 = 2 ** 28
    N5 = 2 ** 35
    N6 = 2 ** 42
    N7 = 2 ** 49
    N8 = 2 ** 56
    N9 = 2 ** 63

    if value < N1:
        return 1
    elif value < N2:
        return 2
    elif value < N3:
        return 3
    elif value < N4:
        return 4
    elif value < N5:
        return 5
    elif value < N6:
        return 6
    elif value < N7:
        return 7
    elif value < N8:
        return 8
    elif value < N9:
        return 9
    else:
        return 10



from abc import ABC


class MightyClient(ABC):
    def get_info(self):
        # TODO: Implement return Mighty Info from the server
        pass

class MightyDefaultClient(MightyClient):
    def __init__(self):
        pass

    def get_data(self) -> str:
        raise NotImplementedError("Not available")
    



from .utils import concat_uint8_arrays, varint_prefixed, utf8_string_to_bytes, utf8_bytes_to_string, extract_varint_prefixed

METADATA_VERSION = "0.1.0"

class AADMeta:
    def __init__(self, version: str, type: str, nonce: bytearray) -> None:
        """
        Initialize an AADMeta instance.

        :param version: The version of the encrypted object.
        :param type: The type name of the encrypted object.
        :param nonce: The nonce used for the encryption scheme.
        """
        self.version = version
        self.type = type
        self.nonce = nonce
        self.METADATA_VERSION = "0.1.0"

    def serialize(self):
        """
        Serialize the AAD header.

        A serialized AAD header contains four pieces of information:
        - version of the metadata format
        - version of the encrypted object
        - type name of the encrypted object
        - nonce used for the encryption scheme

        It is composed of several varint-prefixed Uint8Arrays, which is then itself expressed as a
        varint-prefixed byte array.

        It looks like this on the wire:
        NNxxxxxxxxxxxxxxxxxxxxxxxxx...
          AAxx...BBxx...CCxx...DDxx...

        where AA, BB, CC, DD, and NN are varint-encoded (1-10 bytes long) and express the number of bytes following
        that indicator which comprise that field.

        - AAxxx is the prefixed metadata format version
        - BBxxx is the prefixed object version
        - CCxxx is the prefixed typename
        - DDxxx is the prefixed nonce. Length is prefixed instead of static to allow for multiple envelope types.
        
        - NNxxx is the prefixed length of those four strings concatenated together.

        :return: A varint-prefixed byte array representing the serialized AAD header.
        """
        data = concat_uint8_arrays(
            varint_prefixed(utf8_string_to_bytes(METADATA_VERSION)),
            varint_prefixed(utf8_string_to_bytes(self.version)),
            varint_prefixed(utf8_string_to_bytes(self.type)),
            varint_prefixed(self.nonce)
        )

        return varint_prefixed(data)
    
    def deserialize(data):
        """
        Deserialize the given data into an AADMeta instance.

        :param data: The data to be deserialized.
        :return: A dictionary containing the deserialized metadata, raw metadata, and content.
        :raises Exception: If the metadata version is unrecognized or if there is unexpected additional content in the header.
        """
        header = extract_varint_prefixed({"bs": data})

        raw_metadata = varint_prefixed(header)
        content = data[len(raw_metadata):]

        header_buf = {"bs": header}

        metadata_version = utf8_bytes_to_string(extract_varint_prefixed(header_buf))
        if (metadata_version != METADATA_VERSION):
            raise Exception("Unrecognized metadata version")
        
        metadata = AADMeta(
            version = utf8_bytes_to_string(extract_varint_prefixed(header_buf)),
            type = utf8_bytes_to_string(extract_varint_prefixed(header_buf)),
            nonce = extract_varint_prefixed(header_buf)
        )

        if (len(header_buf["bs"]) != 0):
            raise Exception("Unexpected additional content in header")
        
        return {
            "metadata": metadata,
            "raw_metadata": raw_metadata,
            "content": content
        }



import json
from .utils import utf8_string_to_bytes, utf8_bytes_to_string
from semver_range import Range

class JSONDatagram:
    """
    A class to handle JSON datagrams with versioning and type information.

    Attributes:
        type (str): The type of the datagram.
        version (str): The version of the datagram.
        version_constraint (Range): The version constraint for the datagram.
    """
    def __init__(self, type: str, version: str = "0.0.1", version_constraint: Range = Range("<=0.1.0")):
        """
        Initialize a JSONDatagram instance.

        :param type: The type of the datagram.
        :param version: The version of the datagram.
        :param version_constraint: The version constraint for the datagram.
        """
        # The version constraint is not used (currently)
        self.type = type
        self.version = version
        self.version_constraint = version_constraint

    def serialize(self, data) -> bytes:
        """
        Serialize the datagram to a JSON byte array.

        :param data: The data to be serialized.
        :return: The serialized data as a byte array.
        """
        return utf8_string_to_bytes(json.dumps({"version":self.version, "type": self.type, "data": data},separators=(',', ':')))
        
    def deserialize(self, data: bytes) -> str:
        """
        Deserialize the JSON byte array to retrieve the data.

        :param data: The byte array to be deserialized.
        :return: The deserialized data as a string.
        """
        parsed = json.loads(utf8_bytes_to_string(data))

        return parsed["data"]